@@ -191,14 +191,6 @@ int check_fs_bitmaps(char *name)
 	return 0;
 }
 
-char *inode_time_to_string(__u32 xtime, __u32 xtime_extra)
-{
-	__s64 t = (__s32) xtime;
-
-	t += (__s64) (xtime_extra & EXT4_EPOCH_MASK) << 32;
-	return time_to_string(t);
-}
-
 /*
  * This function takes a __s64 time value and converts it to a string,
  * using ctime
@@ -210,7 +202,7 @@ char *time_to_string(__s64 cl)
 	const char	*tz;
 
 	if (do_gmt == -1) {
-		/* The diet libc doesn't respect the TZ environemnt variable */
+		/* The diet libc doesn't respect the TZ environment variable */
 		tz = ss_safe_getenv("TZ");
 		if (!tz)
 			tz = "";
@@ -355,7 +347,7 @@ int strtoblk(const char *cmd, const char *str, const char *errmsg,
  * This is a common helper function used by the command processing
  * routines
  */
-int common_args_process(int argc, char *argv[], int min_argc, int max_argc,
+int common_args_process(int argc, ss_argv_t argv, int min_argc, int max_argc,
 			const char *cmd, const char *usage, int flags)
 {
 	if (argc < min_argc || argc > max_argc) {
@@ -381,7 +373,7 @@ int common_args_process(int argc, char *argv[], int min_argc, int max_argc,
  * do_testi, etc.  Basically, any command which takes a single
  * argument which is a file/inode number specifier.
  */
-int common_inode_args_process(int argc, char *argv[],
+int common_inode_args_process(int argc, ss_argv_t argv,
 			      ext2_ino_t *inode, int flags)
 {
 	if (common_args_process(argc, argv, 2, 2, argv[0], "<file>", flags))
@@ -396,7 +388,7 @@ int common_inode_args_process(int argc, char *argv[],
 /*
  * This is a helper function used by do_freeb, do_setb, and do_testb
  */
-int common_block_args_process(int argc, char *argv[],
+int common_block_args_process(int argc, ss_argv_t argv,
 			      blk64_t *block, blk64_t *count)
 {
 	int	err;
@@ -420,12 +412,12 @@ int common_block_args_process(int argc, char *argv[],
 	return 0;
 }
 
-int debugfs_read_inode_full(ext2_ino_t ino, struct ext2_inode * inode,
-			const char *cmd, int bufsize)
+int debugfs_read_inode2(ext2_ino_t ino, struct ext2_inode * inode,
+			const char *cmd, int bufsize, int flags)
 {
 	int retval;
 
-	retval = ext2fs_read_inode_full(current_fs, ino, inode, bufsize);
+	retval = ext2fs_read_inode2(current_fs, ino, inode, bufsize, flags);
 	if (retval) {
 		com_err(cmd, retval, "while reading inode %u", ino);
 		return 1;
@@ -446,15 +438,14 @@ int debugfs_read_inode(ext2_ino_t ino, struct ext2_inode * inode,
 	return 0;
 }
 
-int debugfs_write_inode_full(ext2_ino_t ino,
-			     struct ext2_inode *inode,
-			     const char *cmd,
-			     int bufsize)
+int debugfs_write_inode2(ext2_ino_t ino,
+			 struct ext2_inode *inode,
+			 const char *cmd,
+			 int bufsize, int flags)
 {
 	int retval;
 
-	retval = ext2fs_write_inode_full(current_fs, ino,
-					 inode, bufsize);
+	retval = ext2fs_write_inode2(current_fs, ino, inode, bufsize, flags);
 	if (retval) {
 		com_err(cmd, retval, "while writing inode %u", ino);
 		return 1;
@@ -522,7 +513,7 @@ errcode_t read_list(char *str, blk64_t **list, size_t *len)
 	blk64_t *lst = *list;
 	size_t ln = *len;
 	char *tok, *p = str;
-	errcode_t retval;
+	errcode_t retval = 0;
 
 	while ((tok = strtok(p, ","))) {
 		blk64_t *l;
@@ -531,24 +522,28 @@ errcode_t read_list(char *str, blk64_t **list, size_t *len)
 
 		errno = 0;
 		y = x = strtoull(tok, &e, 0);
-		if (errno)
-			return errno;
+		if (errno) {
+			retval = errno;
+			break;
+		}
 		if (*e == '-') {
 			y = strtoull(e + 1, NULL, 0);
-			if (errno)
-				return errno;
+			if (errno) {
+				retval = errno;
+				break;
+			}
 		} else if (*e != 0) {
 			retval = EINVAL;
-			goto err;
+			break;
 		}
 		if (y < x) {
 			retval = EINVAL;
-			goto err;
+			break;
 		}
 		l = realloc(lst, sizeof(blk64_t) * (ln + y - x + 1));
 		if (l == NULL) {
 			retval = ENOMEM;
-			goto err;
+			break;
 		}
 		lst = l;
 		for (; x <= y; x++)
@@ -558,8 +553,40 @@ errcode_t read_list(char *str, blk64_t **list, size_t *len)
 
 	*list = lst;
 	*len = ln;
-	return 0;
-err:
-	free(lst);
 	return retval;
 }
+
+void do_byte_hexdump(FILE *fp, unsigned char *buf, size_t bufsize)
+{
+	size_t		i, j, max;
+	int		suppress = -1;
+
+	for (i = 0; i < bufsize; i += 16) {
+		max = (bufsize - i > 16) ? 16 : bufsize - i;
+		if (suppress < 0) {
+			if (i && memcmp(buf + i, buf + i - max, max) == 0) {
+				suppress = i;
+				fprintf(fp, "*\n");
+				continue;
+			}
+		} else {
+			if (memcmp(buf + i, buf + suppress, max) == 0)
+				continue;
+			suppress = -1;
+		}
+		fprintf(fp, "%04o  ", (unsigned int)i);
+		for (j = 0; j < 16; j++) {
+			if (j < max)
+				fprintf(fp, "%02x", buf[i+j]);
+			else
+				fprintf(fp, "  ");
+			if ((j % 2) == 1)
+				fprintf(fp, " ");
+		}
+		fprintf(fp, " ");
+		for (j = 0; j < max; j++)
+			fprintf(fp, "%c", isprint(buf[i+j]) ? buf[i+j] : '.');
+		fprintf(fp, "\n");
+	}
+	fprintf(fp, "\n");
+}
