@@ -31,6 +31,7 @@ static errcode_t e2fsck_handle_read_error(io_channel channel,
 	int	i;
 	char	*p;
 	ext2_filsys fs = (ext2_filsys) channel->app_data;
+	errcode_t retval;
 	e2fsck_t ctx;
 
 	ctx = (e2fsck_t) fs->priv_data;
@@ -58,9 +59,19 @@ static errcode_t e2fsck_handle_read_error(io_channel channel,
 		printf(_("Error reading block %lu (%s).  "), block,
 		       error_message(error));
 	preenhalt(ctx);
+
+	/* Don't rewrite a block past the end of the FS. */
+	if (block >= ext2fs_blocks_count(fs->super))
+		return 0;
+
 	if (ask(ctx, _("Ignore error"), 1)) {
-		if (ask(ctx, _("Force rewrite"), 1))
-			io_channel_write_blk64(channel, block, count, data);
+		if (ask(ctx, _("Force rewrite"), 1)) {
+			retval = io_channel_write_blk64(channel, block,
+							count, data);
+			if (retval)
+				printf(_("Error rewriting block %lu (%s)\n"),
+				       block, error_message(retval));
+		}
 		return 0;
 	}
 
