@@ -26,6 +26,13 @@ static inline int ext2fsP_is_disk_device(mode_t mode)
 #endif
 }
 
+static inline time_t ext2fsP_get_time(ext2_filsys fs)
+{
+	if (fs->now || (fs->flags2 & EXT2_FLAG2_USE_FAKE_TIME))
+		return fs->now;
+	return time(NULL);
+}
+
 /*
  * Badblocks list
  */
@@ -92,6 +99,25 @@ struct ext2_inode_cache_ent {
 	struct ext2_inode	*inode;
 };
 
+/*
+ * NLS definitions
+ */
+struct ext2fs_nls_table {
+	int version;
+	const struct ext2fs_nls_ops *ops;
+};
+
+struct ext2fs_nls_ops {
+	int (*casefold)(const struct ext2fs_nls_table *charset,
+			const unsigned char *str, size_t len,
+			unsigned char *dest, size_t dlen);
+	int (*validate)(const struct ext2fs_nls_table *table,
+			char *s, size_t len, char **pos);
+	int (*casefold_cmp)(const struct ext2fs_nls_table *table,
+			    const unsigned char *str1, size_t len1,
+			    const unsigned char *str2, size_t len2);
+};
+
 /* Function prototypes */
 
 extern int ext2fs_process_dir_block(ext2_filsys  	fs,
@@ -188,4 +214,7 @@ typedef void (*ext2_exit_fn)(void *);
 errcode_t ext2fs_add_exit_fn(ext2_exit_fn fn, void *data);
 errcode_t ext2fs_remove_exit_fn(ext2_exit_fn fn, void *data);
 
+#define ARRAY_SIZE(array)			\
+        (sizeof(array) / sizeof(array[0]))
+
 #define EXT2FS_BUILD_BUG_ON(cond) ((void)sizeof(char[1 - 2*!!(cond)]))
