@@ -98,6 +98,11 @@ static inline struct page * rb_insert_page_cache(struct inode * inode,
 #include <stdint.h>
 #include "compiler.h"
 
+#if __GNUC_PREREQ (4, 6)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunused-function"
+#endif
+
 struct rb_node
 {
 	uintptr_t  rb_parent_color;
@@ -158,7 +163,7 @@ extern struct rb_node *ext2fs_rb_first(const struct rb_root *);
 extern struct rb_node *ext2fs_rb_last(const struct rb_root *);
 
 /* Fast replacement of a single node without remove/rebalance/add/rebalance */
-extern void ext2fs_rb_replace_node(struct rb_node *victim, struct rb_node *new,
+extern void ext2fs_rb_replace_node(struct rb_node *victim, struct rb_node *new_,
 				 struct rb_root *root);
 
 static inline void ext2fs_rb_link_node(struct rb_node * node,
@@ -171,4 +176,8 @@ static inline void ext2fs_rb_link_node(struct rb_node * node,
 	*rb_link = node;
 }
 
+#if __GNUC_PREREQ (4, 6)
+#pragma GCC diagnostic pop
+#endif
+
 #endif	/* _LINUX_RBTREE_H */
