@@ -97,7 +97,7 @@ static unsigned int sys_page_size = 4096;
 static void usage(void)
 {
 	fprintf(stderr, _(
-"Usage: %s [-b block_size] [-i input_file] [-o output_file] [-svwnf]\n"
+"Usage: %s [-b block_size] [-i input_file] [-o output_file] [-svwnfBX]\n"
 "       [-c blocks_at_once] [-d delay_factor_between_reads] [-e max_bad_blocks]\n"
 "       [-p num_passes] [-t test_pattern [-t test_pattern [...]]]\n"
 "       device [last_block [first_block]]\n"),
@@ -130,7 +130,7 @@ static void *allocate_buffer(size_t size)
 	void	*ret = 0;
 
 #ifdef HAVE_POSIX_MEMALIGN
-	if (posix_memalign(&ret, sys_page_size, size) < 0)
+	if (posix_memalign(&ret, sys_page_size, size) != 0)
 		ret = 0;
 #else
 #ifdef HAVE_MEMALIGN
@@ -389,7 +389,7 @@ static int do_read (int dev, unsigned char * buffer, int try, int block_size,
 	/* Try the read */
 	if (d_flag)
 		gettimeofday(&tv1, NULL);
-	got = read (dev, buffer, try * block_size);
+	got = read (dev, buffer, (size_t) try * block_size);
 	if (d_flag)
 		gettimeofday(&tv2, NULL);
 	if (got < 0)
@@ -460,7 +460,7 @@ static int do_write(int dev, unsigned char * buffer, int try, int block_size,
 		com_err (program_name, errno, "%s", _("during seek"));
 
 	/* Try the write */
-	got = write (dev, buffer, try * block_size);
+	got = write (dev, buffer, (size_t) try * block_size);
 	if (got < 0)
 		got = 0;
 	if (got & 511)
@@ -510,9 +510,9 @@ static unsigned int test_ro (int dev, blk_t last_block,
 	} while (next_bad && next_bad < first_block);
 
 	if (t_flag) {
-		blkbuf = allocate_buffer((blocks_at_once + 1) * block_size);
+		blkbuf = allocate_buffer(((size_t) blocks_at_once + 1) * block_size);
 	} else {
-		blkbuf = allocate_buffer(blocks_at_once * block_size);
+		blkbuf = allocate_buffer((size_t) blocks_at_once * block_size);
 	}
 	if (!blkbuf)
 	{
@@ -612,7 +612,7 @@ static unsigned int test_rw (int dev, blk_t last_block,
 	/* set up abend handler */
 	capture_terminate(NULL);
 
-	buffer = allocate_buffer(2 * blocks_at_once * block_size);
+	buffer = allocate_buffer((size_t) 2 * blocks_at_once * block_size);
 	read_buffer = buffer + blocks_at_once * block_size;
 
 	if (!buffer) {
@@ -771,7 +771,7 @@ static unsigned int test_nd (int dev, blk_t last_block,
 		ext2fs_badblocks_list_iterate (bb_iter, &next_bad);
 	} while (next_bad && next_bad < first_block);
 
-	blkbuf = allocate_buffer(3 * blocks_at_once * block_size);
+	blkbuf = allocate_buffer((size_t) 3 * blocks_at_once * block_size);
 	test_record = malloc(blocks_at_once * sizeof(struct saved_blk_record));
 	if (!blkbuf || !test_record) {
 		com_err(program_name, ENOMEM, "%s",
@@ -892,7 +892,6 @@ static unsigned int test_nd (int dev, blk_t last_block,
 			test_ptr += got * block_size;
 			currently_testing += got;
 			if (got != try) {
-				try = 1;
 				if (recover_block == ~0U)
 					recover_block = currently_testing -
 						got + blocks_at_once;
@@ -1037,10 +1036,13 @@ static unsigned int parse_uint(const char *str, const char *descr)
 
 	errno = 0;
 	ret = strtoul(str, &tmp, 0);
-	if (*tmp || errno || (ret > UINT_MAX) ||
-	    (ret == ULONG_MAX && errno == ERANGE)) {
+	if (*tmp || errno) {
 		com_err (program_name, 0, _("invalid %s - %s"), descr, str);
 		exit (1);
+	} else if ((ret > UINT_MAX) ||
+	    (ret == ULONG_MAX && errno == ERANGE)) {
+		com_err (program_name, 0, _("%s too large - %lu"), descr, ret);
+		exit (1);
 	}
 	return ret;
 }
@@ -1053,7 +1055,7 @@ int main (int argc, char ** argv)
 	char * input_file = NULL;
 	char * output_file = NULL;
 	FILE * in = NULL;
-	int block_size = 1024;
+	unsigned int block_size = 1024;
 	unsigned int blocks_at_once = 64;
 	blk64_t last_block, first_block;
 	int num_passes = 0;
@@ -1066,7 +1068,7 @@ int main (int argc, char ** argv)
 				  unsigned int);
 	int open_flag;
 	long sysval;
-	blk64_t inblk;
+	unsigned long long inblk;
 
 	setbuf(stdout, NULL);
 	setbuf(stderr, NULL);
@@ -1094,6 +1096,8 @@ int main (int argc, char ** argv)
 
 	if (argc && *argv)
 		program_name = *argv;
+	else
+		usage();
 	while ((c = getopt (argc, argv, "b:d:e:fi:o:svwnc:p:h:t:BX")) != EOF) {
 		switch (c) {
 		case 'b':
@@ -1115,13 +1119,13 @@ int main (int argc, char ** argv)
 			v_flag++;
 			break;
 		case 'w':
-			if (w_flag)
+			if (w_flag && w_flag != 1)
 				exclusive_usage();
 			test_func = test_rw;
 			w_flag = 1;
 			break;
 		case 'n':
-			if (w_flag)
+			if (w_flag && w_flag != 2)
 				exclusive_usage();
 			test_func = test_nd;
 			w_flag = 2;
@@ -1201,12 +1205,29 @@ int main (int argc, char ** argv)
 			exit(1);
 		}
 	}
+	if ((block_size == 0) || (block_size > (1 << 24)) ||
+	    (block_size & (block_size - 1))) {
+		com_err(program_name, 0, _("Invalid block size: %u\n"),
+			block_size);
+		exit(1);
+	}
+	if (blocks_at_once == 0) {
+		com_err(program_name, 0, _("Invalid number of blocks: %d\n"),
+			blocks_at_once);
+		exit(1);
+	} else if (((size_t) block_size * blocks_at_once) > SIZE_MAX / 3) {
+		/* maximum usage is in test_nd() */
+		com_err(program_name, 0, _("For block size %d, number of blocks too large: %d\n"),
+			block_size, blocks_at_once);
+		exit(1);
+	}
+
 	if (optind > argc - 1)
 		usage();
 	device_name = argv[optind++];
 	if (optind > argc - 1) {
 		errcode = ext2fs_get_device_size2(device_name,
-						 block_size,
+						 (int) block_size,
 						 &last_block);
 		if (errcode == EXT2_ET_UNIMPLEMENTED) {
 			com_err(program_name, 0, "%s",
@@ -1231,14 +1252,15 @@ int main (int argc, char ** argv)
 	} else first_block = 0;
 	if (first_block >= last_block) {
 	    com_err (program_name, 0, _("invalid starting block (%llu): must be less than %llu"),
-		     first_block, last_block);
+		     (unsigned long long) first_block,
+		     (unsigned long long) last_block);
 	    exit (1);
 	}
 	/* ext2 badblocks file can't handle large values */
 	if (last_block >> 32) {
 		com_err(program_name, EOVERFLOW,
 			_("invalid end block (%llu): must be 32-bit value"),
-			last_block);
+			(unsigned long long) last_block);
 		exit(1);
 	}
 	if (w_flag)
@@ -1335,7 +1357,7 @@ int main (int argc, char ** argv)
 	do {
 		unsigned int bb_count;
 
-		bb_count = test_func(dev, last_block, block_size,
+		bb_count = test_func(dev, last_block, (int) block_size,
 				     first_block, blocks_at_once);
 		if (bb_count)
 			passes_clean = 0;
